# Copyright (c) 2017 "Shopify inc." All rights reserved.
# Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
from __future__ import unicode_literals

import collections
import datetime
import re
import sys

import enum
import typing  # pylint: disable=unused-import
import untangle

from pyoozie import exceptions


_COORD_ID_RE = re.compile('^(?P<id>.*-C)(?:@(?P<action>[1-9][0-9]*))?$')
_WORKFLOW_ID_RE = re.compile('^(?P<id>.*-W)(?:@(?P<action>.*))?$')


def parse_coordinator_id(string):
    parts = _COORD_ID_RE.match(string) if string else None
    coord_id = parts.group('id') if parts else None
    action = parts.group('action') if parts else None
    action = int(action) if action else None
    return coord_id, action


def parse_workflow_id(string):
    parts = _WORKFLOW_ID_RE.match(string) if string else None
    coord_id = parts.group('id') if parts else None
    action = parts.group('action') if parts else None
    return coord_id, action


def _parse_coordinator_id(_, job_id):
    if job_id:
        coord_id, action = parse_coordinator_id(job_id)
        if coord_id and not action:
            return job_id
        raise exceptions.OozieException.parse_error("Invalid coordinator id: {}".format(job_id))
    return None


def _parse_coordinator_action_id(_, job_id):
    if job_id:
        coord_id, action = parse_coordinator_id(job_id)
        if coord_id and action:
            return job_id
        raise exceptions.OozieException.parse_error("Invalid coordinator action id: {}".format(job_id))
    return None


def _parse_workflow_id(_, job_id):
    if job_id:
        wf_id, action = parse_workflow_id(job_id)
        if wf_id and not action:
            return job_id
        raise exceptions.OozieException.parse_error("Invalid workflow id: {}".format(job_id))
    return None


def _parse_workflow_action_id(_, job_id):
    if job_id:
        wf_id, action = parse_workflow_id(job_id)
        if wf_id and action:
            return job_id
        raise exceptions.OozieException.parse_error("Invalid workflow action id: {}".format(job_id))
    return None


def _parse_workflow_parent_id(_, job_id):
    if job_id:
        wf_id, action = parse_coordinator_id(job_id)
        if wf_id and action:
            return job_id
        wf_id, action = parse_workflow_id(job_id)
        if wf_id and not action:
            return job_id
        raise exceptions.OozieException.parse_error("Invalid workflow parent id: {}".format(job_id))
    return None


def _parse_time(_, time_string):
    if time_string:
        try:
            return datetime.datetime.strptime(time_string, '%a, %d %b %Y %H:%M:%S %Z')
        except ValueError as err:
            raise exceptions.OozieException.parse_error("Error parsing time '{}'".format(time_string), err)
    return None


def _parse_configuration(_, conf_string):
    if conf_string is None:
        return None
    elif conf_string:
        xml = conf_string if sys.version_info >= (3, 0) else conf_string.encode('utf-8')
        conf = untangle.parse(xml).configuration
        return {prop.name.cdata: prop.value.cdata for prop in conf.property}
    else:
        return {}


def _parse_workflow_actions(artifact, actions_list):
    actions_list = actions_list or []
    actions = [WorkflowAction(artifact._client, action, parent=artifact) for action in actions_list]
    return {action.name: action for action in actions}


def _parse_coordinator_actions(artifact, actions_list):
    actions_list = actions_list or []
    actions = [CoordinatorAction(artifact._client, action, parent=artifact) for action in actions_list]
    return {action.actionNumber: action for action in actions}


def _parse_coordinator_status(_, status_string):
    return CoordinatorStatus.parse(status_string)


def _parse_coordinator_action_status(_, status_string):
    return CoordinatorActionStatus.parse(status_string)


def _parse_workflow_status(_, status_string):
    return WorkflowStatus.parse(status_string)


def _parse_workflow_action_status(_, status_string):
    return WorkflowActionStatus.parse(status_string)


class ArtifactType(enum.Enum):
    Coordinator = 1
    CoordinatorAction = 2
    Workflow = 3
    WorkflowAction = 4


_StatusValue = collections.namedtuple('_StatusValue',
                                      ['status_id', 'is_active', 'is_running', 'is_suspendable', 'is_suspended'])


def _status(status_id, is_active=False, is_running=False, is_suspendable=False, is_suspended=False):
    if is_running and not is_active:
        raise exceptions.OozieException.parse_error("A running status implies active")
    return _StatusValue(status_id, is_active, is_running, is_suspendable, is_suspended)


class _StatusEnum(enum.Enum):

    def __str__(self):
        return self.name

    @classmethod
    def parse(cls, status_string):
        values = cls.__members__
        return values.get(status_string, values['UNKNOWN'])

    @classmethod
    def active(cls):
        return [status for status in cls if status.is_active()]

    @classmethod
    def running(cls):
        return [status for status in cls if status.is_running()]

    @classmethod
    def suspendable(cls):
        return [status for status in cls if status.is_suspendable()]

    @classmethod
    def suspended(cls):
        return [status for status in cls if status.is_suspended()]

    def is_unknown(self):
        return self._value_.status_id == 0

    def is_active(self):
        return self._value_.is_active

    def is_running(self):
        return self._value_.is_running

    def is_suspendable(self):
        return self._value_.is_suspendable

    def is_suspended(self):
        return self._value_.is_suspended


class CoordinatorStatus(_StatusEnum):
    UNKNOWN = _status(0)
    DONEWITHERROR = _status(1)
    FAILED = _status(2)
    IGNORED = _status(3)
    KILLED = _status(4)
    PAUSED = _status(5, is_active=True)
    PAUSEDWITHERROR = _status(6, is_active=True)
    PREMATER = _status(7, is_active=True)
    PREP = _status(8, is_active=True, is_suspendable=True)
    PREPPAUSED = _status(9, is_active=True)
    PREPSUSPENDED = _status(10, is_active=True, is_suspended=True)
    RUNNING = _status(11, is_active=True, is_running=True, is_suspendable=True)
    RUNNINGWITHERROR = _status(12, is_active=True, is_running=True, is_suspendable=True)
    SUCCEEDED = _status(13)
    SUSPENDED = _status(14, is_active=True, is_running=True, is_suspended=True)
    SUSPENDEDWITHERROR = _status(15, is_active=True, is_running=True, is_suspended=True)


class CoordinatorActionStatus(_StatusEnum):
    UNKNOWN = _status(0)
    FAILED = _status(1)
    IGNORED = _status(2)
    KILLED = _status(3)
    READY = _status(4, is_active=True)
    RUNNING = _status(5, is_active=True, is_running=True, is_suspendable=True)
    SKIPPED = _status(6)
    SUBMITTED = _status(7, is_active=True)
    SUCCEEDED = _status(8)
    SUSPENDED = _status(9, is_active=True, is_running=True, is_suspended=True)
    TIMEDOUT = _status(10)
    WAITING = _status(11, is_active=True)


class WorkflowStatus(_StatusEnum):
    UNKNOWN = _status(0)
    FAILED = _status(1)
    KILLED = _status(2)
    PREP = _status(3, is_active=True)
    RUNNING = _status(4, is_active=True, is_running=True, is_suspendable=True)
    SUCCEEDED = _status(5)
    SUSPENDED = _status(6, is_active=True, is_running=True, is_suspended=True)


class WorkflowActionStatus(_StatusEnum):
    UNKNOWN = _status(0)
    DONE = _status(1)
    END_MANUAL = _status(2)
    END_RETRY = _status(3)
    ERROR = _status(4)
    FAILED = _status(5)
    KILLED = _status(6)
    OK = _status(7)
    PREP = _status(8, is_active=True)
    RUNNING = _status(9, is_active=True, is_running=True)
    START_MANUAL = _status(10)
    START_RETRY = _status(11)
    USER_RETRY = _status(12, is_active=True)


class _OozieArtifact(object):

    REQUIRED_KEYS = {}  # type: typing.Dict[unicode, typing.Callable]

    SUPPORTED_KEYS = {'toString': None}  # type: typing.Dict[unicode, typing.Optional[typing.Callable]]

    def __init__(self, oozie_client, details, parent=None):
        self._client = oozie_client
        self._parent = parent
        details = dict(details)
        for key, func in self.REQUIRED_KEYS.items():
            value = details.pop(key, None)
            try:
                parsed_value = func(self, value) if func else value
            except exceptions.OozieException as err:
                raise exceptions.OozieException.required_key_missing(key, self, err)
            if parsed_value is None:
                raise exceptions.OozieException.required_key_missing(key, self)
            else:
                setattr(self, key, parsed_value)
        for key, func in self.SUPPORTED_KEYS.items():
            value = details.pop(key, None)
            value = func(self, value) if func else value
            setattr(self, key, value)
        self._details = details
        self._validate_degenerate_fields()

    def __str__(self):
        return self.toString

    def fill_in_details(self):
        # Fetch any missing data not supplied
        return self

    def _validate_degenerate_fields(self):
        # For any fields that must be in sync, ensure they are.
        # If values are missing, extrapolate them
        pass

    @property
    def client_url(self):
        return self._client.url if self._client else None

    def is_coordinator(self):
        return False

    def is_workflow(self):
        return False

    def is_action(self):
        return False


class Coordinator(_OozieArtifact):

    REQUIRED_KEYS = {
        'coordJobId': _parse_coordinator_id,
    }

    SUPPORTED_KEYS = {
        'acl': None,
        'actions': _parse_coordinator_actions,
        'bundleId': None,
        'concurrency': None,
        'conf': _parse_configuration,
        'consoleUrl': None,
        'coordExternalId': None,
        'coordJobName': None,
        'coordJobPath': None,
        'endTime': _parse_time,
        'executionPolicy': None,
        'frequency': None,
        'group': None,
        'lastAction': _parse_time,
        'mat_throttling': None,
        'nextMaterializedTime': _parse_time,
        'pauseTime': None,
        'startTime': _parse_time,
        'status': _parse_coordinator_status,
        'timeOut': None,
        'timeUnit': None,
        'timeZone': None,
        'toString': None,
        'total': None,
        'user': None,
    }

    def __init__(self, *args, **kwargs):
        super(Coordinator, self).__init__(*args, **kwargs)
        self._workflow = None

    def fill_in_details(self):
        # Undefined `conf` is probably bad, empty is ok
        if self.conf is None:
            coord = self._client.job_last_coordinator_info(coordinator_id=self.coordJobId)
            return coord
        else:
            return self

    def _validate_degenerate_fields(self):
        # For any fields that must be in sync, ensure they are.
        # If values are missing, extrapolate them
        self.toString = self.toString or 'Coordinator application id[{}] status[{}]'.format(
            self.coordJobId,
            self.status)
        if self.coordJobId not in self.toString:
            raise exceptions.OozieException.parse_error("toString does not contain coordinator ID")
        if not self.status.is_unknown() and str(self.status) not in self.toString:
            raise exceptions.OozieException.parse_error("toString does not contain status")

    def is_coordinator(self):
        return True

    def coordinator(self):
        return self

    def parent(self):
        return None

    def action(self, number):
        if number in self.actions:
            action = self.actions[number]
        else:
            action = self._client.job_coordinator_action(action_number=number, coordinator=self)
        return action


class CoordinatorAction(_OozieArtifact):

    REQUIRED_KEYS = {
        'id': _parse_coordinator_action_id,
    }

    SUPPORTED_KEYS = {
        'actionNumber': None,
        'consoleUrl': None,
        'coordJobId': _parse_coordinator_id,
        'createdConf': None,
        'createdTime': _parse_time,
        'errorCode': None,
        'errorMessage': None,
        'externalId': _parse_workflow_id,
        'externalStatus': None,
        'lastModifiedTime': _parse_time,
        'missingDependencies': None,
        'nominalTime': _parse_time,
        'pushMissingDependencies': None,
        'runConf': None,
        'status': _parse_coordinator_action_status,
        'toString': None,
        'trackerUri': None,
        'type': None,
    }

    def __init__(self, *args, **kwargs):
        self.status = CoordinatorActionStatus.UNKNOWN
        super(CoordinatorAction, self).__init__(*args, **kwargs)
        self._workflow = None

    def _validate_degenerate_fields(self):
        # For any fields that must be in sync, ensure they are.
        # If values are missing, extrapolate them
        coord_id, action = parse_coordinator_id(self.id)
        self.coordJobId = self.coordJobId or coord_id
        if self.coordJobId != coord_id:
            raise exceptions.OozieException.parse_error("coordJobId does not match coordinator action ID")
        self.actionNumber = self.actionNumber or action
        if self.actionNumber != action:
            raise exceptions.OozieException.parse_error("actionNumber does not match coordinator action ID")
        self.toString = self.toString or 'CoordinatorAction name[{}] status[{}]'.format(self.id, self.status)
        if self.id not in self.toString:
            raise exceptions.OozieException.parse_error("toString does not contain coordinator action ID")
        if not self.status.is_unknown() and str(self.status) not in self.toString:
            raise exceptions.OozieException.parse_error("toString does not contain status")

    def is_coordinator(self):
        return True

    def is_action(self):
        return True

    def workflow(self):
        # TODO: revisit this to support multiple runs
        # Use .../job/...-C@xx?show=allruns to query
        if not self._workflow and self.externalId:
            workflow = self._client.job_workflow_info(workflow_id=self.externalId)
            if workflow:
                workflow._parent = self
                self._workflow = workflow
        return self._workflow

    def coordinator(self):
        if not self._parent:
            self._parent = self._client.job_coordinator_info(coordinator_id=self.coordJobId)
        return self._parent

    def coordinator_action(self):
        return self

    def parent(self):
        return self.coordinator()


class Workflow(_OozieArtifact):

    REQUIRED_KEYS = {
        'id': _parse_workflow_id,
    }

    SUPPORTED_KEYS = {
        'acl': None,
        'actions': _parse_workflow_actions,
        'appName': None,
        'appPath': None,
        'conf': _parse_configuration,
        'consoleUrl': None,
        'createdTime': _parse_time,
        'endTime': _parse_time,
        'externalId': None,
        'group': None,
        'lastModTime': _parse_time,
        'parentId': _parse_workflow_parent_id,
        'run': None,
        'startTime': _parse_time,
        'status': _parse_workflow_status,
        'toString': None,
        'user': None,
    }

    def __init__(self, *args, **kwargs):
        super(Workflow, self).__init__(*args, **kwargs)
        self._workflow = None

    def fill_in_details(self):
        # Undefined `conf` is probably bad, empty is ok
        if self.conf is None:
            workflow = self._client.job_workflow_info(workflow_id=self.id)
            return workflow
        else:
            return self

    def _validate_degenerate_fields(self):
        # For any fields that must be in sync, ensure they are.
        # If values are missing, extrapolate them
        self.toString = self.toString or 'Workflow id[{}] status[{}]'.format(self.id, self.status)
        if self.id not in self.toString:
            raise exceptions.OozieException.parse_error("toString does not contain workflow ID")
        if not self.status.is_unknown() and str(self.status) not in self.toString:
            raise exceptions.OozieException.parse_error("toString does not contain status")

    def is_workflow(self):
        return True

    def coordinator(self):
        parent = self.parent()
        if parent:
            return parent.coordinator()

    def coordinator_action(self):
        parent = self.parent()
        if parent:
            return parent.coordinator_action()

    def parent(self):
        if not self._parent and self.parentId:
            self._parent = self._client.job_action_info(self.parentId)
        return self._parent

    def action(self, name):
        return self.actions.get(name, None)


class WorkflowAction(_OozieArtifact):

    REQUIRED_KEYS = {
        'id': _parse_workflow_action_id,
    }

    SUPPORTED_KEYS = {
        'conf': None,
        'consoleUrl': None,
        'cred': None,
        'data': None,
        'endTime': _parse_time,
        'errorCode': None,
        'errorMessage': None,
        'externalChildIDs': None,
        'externalId': None,
        'externalStatus': None,
        'name': None,
        'retries': None,
        'startTime': _parse_time,
        'stats': None,
        'status': _parse_workflow_action_status,
        'toString': None,
        'trackerUri': None,
        'transition': None,
        'type': None,
        'userRetryCount': None,
        'userRetryInterval': None,
        'userRetryMax': None,
    }

    def __init__(self, *args, **kwargs):
        super(WorkflowAction, self).__init__(*args, **kwargs)
        self._subworkflow = None

    def _validate_degenerate_fields(self):
        # For any fields that must be in sync, ensure they are.
        # If values are missing, extrapolate them
        _, action = parse_workflow_id(self.id)
        self.name = self.name or action
        if self.name != action:
            raise exceptions.OozieException.parse_error("name does not match workflow action ID")
        self.toString = self.toString or 'Action name[{}] status[{}]'.format(self.name, self.status)
        if self.name not in self.toString:
            raise exceptions.OozieException.parse_error("toString does not contain workflow action name")
        if not self.status.is_unknown() and str(self.status) not in self.toString:
            raise exceptions.OozieException.parse_error("toString does not contain status")

    def is_workflow(self):
        return True

    def is_action(self):
        return True

    def subworkflow(self):
        if not self._subworkflow and self.type == 'sub-workflow' and self.externalId:
            workflow = self._client.job_workflow_info(self.externalId)
            if workflow:
                workflow._parent = self
                self._subworkflow = workflow
        return self._subworkflow

    def coordinator(self):
        parent = self.parent()
        if parent:
            return parent.coordinator()

    def coordinator_action(self):
        parent = self.parent()
        if parent:
            return parent.coordinator_action()

    def parent(self):
        if not self._parent and self.externalId:
            self._parent = self._client.job_workflow_info(self.externalId)
        return self._parent



# Copyright (c) 2017 "Shopify inc." All rights reserved.
# Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
from __future__ import unicode_literals

import abc
import collections
import copy
import datetime
import enum
import itertools
import re
import string  # pylint: disable=deprecated-module
import typing  # pylint: disable=unused-import
import uuid

import six
import yattag

MAX_NAME_LENGTH = 255
MAX_IDENTIFIER_LENGTH = 50
REGEX_IDENTIFIER = r'^[a-zA-Z_][\-_a-zA-Z0-9]{0,%i}$' % (MAX_IDENTIFIER_LENGTH - 1)
COMPILED_REGEX_IDENTIFIER = re.compile(REGEX_IDENTIFIER)
ALLOWABLE_NAME_CHARS = set(string.ascii_letters + string.punctuation + string.digits + ' ')
ONE_HUNDRED_YEARS = 100 * 365.24

PropertyValuesType = typing.Dict[typing.Text, typing.Any]
JobXmlFilesType = typing.Iterable[typing.Text]


class ExecutionOrder(enum.Enum):
    """Execution order used for coordinator jobs."""

    FIFO = 'FIFO'
    LAST_ONLY = 'LAST_ONLY'
    LIFO = 'LIFO'
    NONE = 'NONE'

    def __str__(self):
        return self.value


EXEC_FIFO = ExecutionOrder.FIFO
EXEC_LAST_ONLY = ExecutionOrder.LAST_ONLY
EXEC_LIFO = ExecutionOrder.LIFO
EXEC_NONE = ExecutionOrder.NONE


def validate_xml_name(name):
    # type: (typing.Text) -> typing.Text

    assert len(name) <= MAX_NAME_LENGTH, \
        "Name must be less than {max_length} chars long, '{name}' is {length}".format(
            max_length=MAX_NAME_LENGTH,
            name=name,
            length=len(name))

    assert all(c in ALLOWABLE_NAME_CHARS for c in name), \
        "Name must be comprised of printable ASCII characters, '{name}' is not".format(name=name)

    return name


def validate_xml_id(identifier):
    # type: (typing.Text) -> typing.Text

    assert len(identifier) <= MAX_IDENTIFIER_LENGTH, \
        "Identifier must be less than {max_length} chars long, '{identifier}' is {length}".format(
            max_length=MAX_IDENTIFIER_LENGTH,
            identifier=identifier,
            length=len(identifier))

    assert COMPILED_REGEX_IDENTIFIER.match(identifier), \
        "Identifier must match {regex}, '{identifier}' does not".format(
            regex=REGEX_IDENTIFIER,
            identifier=identifier)

    return identifier


class XMLSerializable(object):
    """An abstract object that can be serialized to XML."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, xml_tag):
        # type: (typing.Text) -> None
        self.xml_tag = xml_tag

    def xml(self, indent=False):
        # type: (bool) -> str
        doc, tag, text = yattag.Doc().tagtext()
        doc.asis("<?xml version='1.0' encoding='UTF-8'?>")
        xml = self._xml(doc, tag, text).getvalue()
        if indent:
            xml = yattag.indent(xml, indentation=' ' * 4, newline='\r\n')
        return xml.encode('utf-8')

    @abc.abstractmethod
    def _xml(self, doc, tag, text):
        # type: (yattag.doc.Doc, yattag.doc.Doc.tag, yattag.doc.Doc.text) -> yattag.doc.Doc
        raise NotImplementedError()


class _PropertyList(XMLSerializable, dict):
    """
    Object used to represent Oozie workflow/coordinator property-value sets.

    Generates XML of the form:
    ...
    <xml_tag>
      <property>
        <name>[PROPERTY-NAME]</name>
        <value>[PROPERTY-VALUE]</value>
      </property>
      ...
    </xml_tag>
    """

    def __init__(
            self,
            xml_tag,          # type: typing.Text
            attributes=None,  # type: typing.Optional[typing.Dict[typing.Text, typing.Text]]
            values=None       # type: typing.Optional[PropertyValuesType]
    ):
        # type: (...) -> None
        super(_PropertyList, self).__init__(xml_tag=xml_tag)
        if values:
            self.update(values)
        self.attributes = attributes or {}

    def _xml(self, doc, tag, text):
        # type: (yattag.doc.Doc, yattag.doc.Doc.tag, yattag.doc.Doc.text) -> yattag.doc.Doc
        with tag(self.xml_tag, **self.attributes):
            for name, value in sorted(self.items()):
                with tag('property'):
                    with tag('name'):
                        doc.text('{}'.format(name))
                    with tag('value'):
                        doc.text('{}'.format(value) if value is not None else '')
        return doc


class Parameters(_PropertyList):
    """Coordinator/workflow parameters.

    Allows one to specify properties that can be reused in actions. "Properties that are a valid Java identifier,
    `[A-Za-z_][0-9A-Za-z_]*` , are available as '${NAME}' variables within the workflow definition."

    "Properties that are not valid Java Identifier, for example 'job.tracker', are available via the
    String wf:conf(String name) function. Valid identifier properties are available via this function as well."
    """

    def __init__(self, values=None):
        # type: (typing.Optional[PropertyValuesType]) -> None
        super(Parameters, self).__init__(xml_tag='parameters', values=values)


class Configuration(_PropertyList):
    """Coordinator job submission, workflow, workflow action configuration XML."""

    def __init__(self, values=None):
        # type: (typing.Optional[PropertyValuesType]) -> None
        super(Configuration, self).__init__(xml_tag='configuration', values=values)


class Credential(_PropertyList):
    """HCatalog, Hive Metastore, HBase, or Hive Server 2 action credentials.

    Generates XML of the form::

       ...
       <credentials>
         <credential name='my-hcat-creds' type='hcat'>
            <property>
               <name>hcat.metastore.uri</name>
               <value>HCAT_URI</value>
            </property>
            ...
         </credential>
        </credentials>
        <action name='pig' cred='my-hcat-creds'>
          <pig>
          ...
    """

    def __init__(
            self,
            values,           # type: PropertyValuesType
            credential_name,  # type: typing.Text
            credential_type   # type: typing.Text
    ):
        # type: (...) -> None
        super(Credential, self).__init__(
            xml_tag='credential',
            attributes={
                'name': credential_name,
                'type': credential_type,
            },
            values=values
        )
        self.name = validate_xml_id(credential_name)


class Shell(XMLSerializable):
    """Workflow shell action (v0.3)."""

    def __init__(
            self,
            exec_command,         # type: typing.Text
            job_tracker=None,     # type: typing.Optional[typing.Text]
            name_node=None,       # type: typing.Optional[typing.Text]
            prepare=None,         # type: typing.Optional[typing.Sequence]
            job_xml_files=None,   # type: typing.Optional[JobXmlFilesType]
            configuration=None,   # type: typing.Optional[PropertyValuesType]
            arguments=None,       # type: typing.Optional[typing.Iterable[typing.Text]]
            env_vars=None,        # type: typing.Optional[PropertyValuesType]
            files=None,           # type: typing.Optional[typing.Iterable[typing.Text]]
            archives=None,        # type: typing.Optional[typing.Iterable[typing.Text]]
            capture_output=False  # type: bool
    ):
        # type: (...) -> None
        super(Shell, self).__init__(xml_tag='shell')
        self.exec_command = exec_command
        self.job_tracker = job_tracker
        self.name_node = name_node
        self.prepare = prepare if prepare else []
        self.job_xml_files = job_xml_files if job_xml_files else []
        self.configuration = Configuration(configuration)
        self.arguments = arguments if arguments else []
        self.env_vars = env_vars if env_vars else {}
        self.files = files if files else []
        self.archives = archives if archives else []
        self.capture_output = capture_output

    def _xml(self, doc, tag, text):
        # type: (yattag.doc.Doc, yattag.doc.Doc.tag, yattag.doc.Doc.text) -> yattag.doc.Doc
        with tag(self.xml_tag, xmlns='uri:oozie:shell-action:0.3'):
            if self.job_tracker:
                with tag('job-tracker'):
                    doc.text(self.job_tracker)

            if self.name_node:
                with tag('name-node'):
                    doc.text(self.name_node)

            if self.prepare:
                raise NotImplementedError("Shell action's prepare has not yet been implemented")

            for xml_file in self.job_xml_files:
                with tag('job-xml'):
                    doc.text(xml_file)

            if self.configuration:
                self.configuration._xml(doc, tag, text)

            with tag('exec'):
                doc.text(self.exec_command)

            for argument in self.arguments:
                with tag('argument'):
                    doc.text(argument)

            for key, value in self.env_vars.items():
                with tag('env-var'):
                    doc.text('{key}={value}'.format(key=key, value=value))

            for filename in self.files:
                with tag('file'):
                    doc.text(filename)

            for archive in self.archives:
                with tag('archive'):
                    doc.text(archive)

            if self.capture_output:
                doc.stag('capture-output')

        return doc


class SubWorkflow(XMLSerializable):
    """Run another workflow defined in another XML file on HDFS.

    An Oozie sub-workflow is an "action [that] runs a child workflow job [...]. The parent workflow job will wait
    until the child workflow job has completed."
    """

    def __init__(
            self,
            app_path,                      # type: str
            propagate_configuration=True,  # type: bool
            configuration=None             # type: typing.Optional[PropertyValuesType]
    ):
        # type: (...) -> None
        super(SubWorkflow, self).__init__(xml_tag='sub-workflow')
        self.app_path = app_path
        self.propagate_configuration = propagate_configuration
        self.configuration = Configuration(configuration)

    def _xml(self, doc, tag, text):
        # type: (yattag.doc.Doc, yattag.doc.Doc.tag, yattag.doc.Doc.text) -> yattag.doc.Doc
        with tag(self.xml_tag):
            with tag('app-path'):
                doc.text(self.app_path)
            if self.propagate_configuration:
                doc.stag('propagate-configuration')
            if self.configuration:
                self.configuration._xml(doc, tag, text)

        return doc


class GlobalConfiguration(XMLSerializable):
    """Global configuration values for all actions in a workflow.

    "Oozie allows a global section to reduce the redundant job-tracker and name-node declarations for each action.
    [...] The global section may contain the job-xml, configuration, job-tracker, or name-node that the user would
    like to set for every action.  If a user then redefines one of these in a specific action node, Oozie will
    update [sic] use the specific declaration instead of the global one for that action."

    "The job-xml element, if present, must refer to a Hadoop JobConf job.xml file bundled in the workflow
    application."
    """

    def __init__(
            self,
            job_tracker=None,    # type: typing.Optional[typing.Text]
            name_node=None,      # type: typing.Optional[typing.Text]
            job_xml_files=None,  # type: typing.Optional[JobXmlFilesType]
            configuration=None,  # type: typing.Optional[PropertyValuesType]
    ):
        # type: (...) -> None
        super(GlobalConfiguration, self).__init__(xml_tag='global')
        self.job_tracker = job_tracker
        self.name_node = name_node
        self.job_xml_files = job_xml_files if job_xml_files else []
        self.configuration = Configuration(configuration)

    def _xml(self, doc, tag, text):
        # type: (yattag.doc.Doc, yattag.doc.Doc.tag, yattag.doc.Doc.text) -> yattag.doc.Doc
        with tag(self.xml_tag):
            if self.job_tracker:
                with tag('job-tracker'):
                    doc.text(self.job_tracker)
            if self.name_node:
                with tag('name-node'):
                    doc.text(self.name_node)
            if self.job_xml_files:
                for xml_file in self.job_xml_files:
                    with tag('job-xml'):
                        doc.text(xml_file)
            if self.configuration:
                self.configuration._xml(doc, tag, text)

        return doc


class Email(XMLSerializable):
    """Email action for use within a workflow."""

    def __init__(
            self,
            to,                 # type: typing.Union[typing.Text, typing.Iterable[typing.Text]]
            subject,            # type: typing.Text
            body,               # type: typing.Text
            cc=None,            # type: typing.Optional[typing.Union[typing.Text, typing.Iterable[typing.Text]]]
            bcc=None,           # type: typing.Optional[typing.Union[typing.Text, typing.Iterable[typing.Text]]]
            content_type=None,  # type: typing.Optional[typing.Text]
            attachments=None    # type: typing.Optional[typing.Text]
    ):
        # type: (...) -> None
        super(Email, self).__init__(xml_tag='email')
        self.to = to
        self.subject = subject
        self.body = body
        self.cc = cc
        self.bcc = bcc
        self.content_type = content_type
        self.attachments = attachments

    def _xml(self, doc, tag, text):
        # type: (yattag.doc.Doc, yattag.doc.Doc.tag, yattag.doc.Doc.text) -> yattag.doc.Doc

        def format_list(strings):
            if hasattr(strings, '__iter__') and not isinstance(strings, six.string_types):
                return ','.join(sorted(strings))
            else:
                return strings

        with tag(self.xml_tag, xmlns='uri:oozie:email-action:0.2'):
            with tag('to'):
                doc.text(format_list(self.to))
            with tag('subject'):
                doc.text(self.subject)
            with tag('body'):
                doc.text(self.body)
            if self.cc:
                with tag('cc'):
                    doc.text(format_list(self.cc))
            if self.bcc:
                with tag('bcc'):
                    doc.text(format_list(self.bcc))
            if self.content_type:
                with tag('content_type'):
                    doc.text(self.content_type)
            if self.attachments:
                with tag('attachment'):
                    doc.text(format_list(self.attachments))

        return doc


class CoordinatorApp(XMLSerializable):

    def __init__(
            self,
            name,                         # type: typing.Text
            workflow_app_path,            # type: typing.Text
            frequency,                    # type: int
            start,                        # type: datetime.datetime
            end=None,                     # type: typing.Optional[datetime.datetime]
            timezone=None,                # type: typing.Optional[typing.Text]
            workflow_configuration=None,  # type: typing.Optional[PropertyValuesType]
            timeout=None,                 # type: typing.Optional[int]
            concurrency=None,             # type: typing.Optional[int]
            execution_order=None,         # type: typing.Optional[ExecutionOrder]
            throttle=None,                # type: typing.Optional[int]
            parameters=None               # type: typing.Optional[PropertyValuesType]
    ):
        # type: (...) -> None
        super(CoordinatorApp, self).__init__(xml_tag='coordinator-app')

        # Compose and validate dates/frequencies
        if end is None:
            end = start + datetime.timedelta(days=ONE_HUNDRED_YEARS)
        assert end > start, "End time ({end}) must be greater than the start time ({start})".format(
            end=CoordinatorApp.__format_datetime(end), start=CoordinatorApp.__format_datetime(start))
        assert frequency >= 5, "Frequency ({frequency} min) must be greater than or equal to 5 min".format(
            frequency=frequency)

        # Coordinator
        self.name = validate_xml_name(name)
        self.frequency = frequency
        self.start = start
        self.end = end
        self.timezone = timezone if timezone else 'UTC'

        # Workflow action
        self.workflow_app_path = workflow_app_path
        self.workflow_configuration = Configuration(workflow_configuration)

        # Controls
        self.timeout = timeout
        self.concurrency = concurrency
        self.execution_order = execution_order
        self.throttle = throttle

        self.parameters = Parameters(parameters)

    @staticmethod
    def __format_datetime(value):  # type: (datetime.datetime) -> typing.Text
        return value.strftime('%Y-%m-%dT%H:%MZ')

    def _xml(self, doc, tag, text):
        # type: (yattag.doc.Doc, yattag.doc.Doc.tag, yattag.doc.Doc.text) -> yattag.doc.Doc

        with tag(self.xml_tag, xmlns="uri:oozie:coordinator:0.4", name=self.name, frequency=str(self.frequency),
                 start=CoordinatorApp.__format_datetime(self.start), end=CoordinatorApp.__format_datetime(self.end),
                 timezone=self.timezone):

            if self.parameters:
                self.parameters._xml(doc, tag, text)

            if self.timeout or self.concurrency or self.execution_order or self.throttle:
                with tag('controls'):
                    if self.timeout:
                        with tag('timeout'):
                            text(str(self.timeout))
                    if self.concurrency:
                        with tag('concurrency'):
                            text(str(self.concurrency))
                    if self.execution_order:
                        with tag('execution'):
                            text(str(self.execution_order))
                    if self.throttle:
                        with tag('throttle'):
                            text(str(self.throttle))

            with tag('action'):
                with tag('workflow'):
                    with tag('app-path'):
                        text(self.workflow_app_path)
                    if self.workflow_configuration:
                        self.workflow_configuration._xml(doc, tag, text)

        return doc


class _AbstractWorkflowEntity(collections.Iterable):
    """An abstract object representing an Oozie workflow action that can be serialized to XML."""
    # pylint: disable=abstract-method

    __metaclass__ = abc.ABCMeta

    def __init__(
            self,
            xml_tag,       # type: typing.Text
            name=None,     # type: typing.Optional[typing.Text]
            on_error=None  # type: typing.Optional[_AbstractWorkflowEntity]
    ):
        # type: (...) -> None
        self.__xml_tag = xml_tag or 'unknown'
        self.__name = name if name else uuid.uuid4().hex[:8]
        self.__on_error = copy.deepcopy(on_error)
        self.__identifier = self.create_identifier(self.__xml_tag)

    def xml_tag(self):
        # type: () -> typing.Text
        return self.__xml_tag

    def identifier(self):
        # type: () -> typing.Text
        return self.__identifier

    def create_identifier(self, xml_tag):
        # type: (typing.Text) -> typing.Text
        return validate_xml_id('{tag}-{name}'.format(tag=xml_tag, name=self.__name))

    def _xml_and_get_on_error(
            self,
            doc,        # type: yattag.doc.Doc
            tag,        # type: yattag.doc.Doc.tag
            text,       # type: yattag.doc.Doc.Text
            on_next,    # type: typing.Text
            on_error    # type: typing.Optional[typing.Text]
    ):
        if self.__on_error:
            self.__on_error._xml(doc, tag, text, on_next, on_error)
        return self.__on_error.identifier() if self.__on_error else (
            on_error if on_error else on_next
        )

    @abc.abstractmethod
    def _xml(
            self,
            doc,        # type: yattag.doc.Doc
            tag,        # type: yattag.doc.Doc.tag
            text,       # type: yattag.doc.Doc.Text
            on_next,    # type: typing.Text
            on_error    # type: typing.Optional[typing.Text]
    ):
        raise NotImplementedError()

    def __iter__(self):
        # type: () -> typing.Generator[_AbstractWorkflowEntity, None, None]
        yield self
        if self.__on_error:
            for action in self.__on_error:
                yield action

    def __bool__(self):  # type: () -> bool
        return bool(list(self))

    def __nonzero__(self):  # type: () -> bool
        return self.__bool__()

    def __repr__(self):  # type: () -> str
        return str('<{_class}({identifier})>'.format(_class=type(self).__name__, identifier=self.identifier()))


class Kill(_AbstractWorkflowEntity):
    """Workflow graph terminal node(s) to end upon to indicate failure."""

    def __init__(self, message, name=None):
        # type: (typing.Text, typing.Optional[typing.Text]) -> None
        super(Kill, self).__init__(xml_tag='kill', name=name)
        self.message = message

    def _xml(
            self,
            doc,        # type: yattag.doc.Doc
            tag,        # type: yattag.doc.Doc.tag
            text,       # type: yattag.doc.Doc.Text
            on_next,    # type: typing.Text
            on_error    # type: typing.Optional[typing.Text]
    ):
        with tag(self.xml_tag(), name=self.identifier()):
            with tag('message'):
                doc.text(self.message)
        return doc


ConcreteAction = typing.Union[Shell, SubWorkflow, Email]


class Action(_AbstractWorkflowEntity):
    """Workflow action nodes carrying concrete actions that perform an action."""

    def __init__(
            self,
            action,               # type: ConcreteAction
            name=None,            # type: typing.Optional[typing.Text]
            credential=None,      # type: typing.Optional[typing.Text]
            retry_max=None,       # type: typing.Optional[int]
            retry_interval=None,  # type: typing.Optional[int]
            on_error=None         # type: typing.Optional[_AbstractWorkflowEntity]
    ):
        # type: (...) -> None
        super(Action, self).__init__(xml_tag='action', name=name, on_error=on_error)

        # XML-document-related values
        self.__action = action
        self.__credential = credential
        self.__retry_max = retry_max
        self.__retry_interval = retry_interval

    def credential(self):
        # type: () -> typing.Optional[typing.Text]
        return self.__credential

    def _xml(
            self,
            doc,        # type: yattag.doc.Doc
            tag,        # type: yattag.doc.Doc.tag
            text,       # type: yattag.doc.Doc.Text
            on_next,    # type: typing.Text
            on_error    # type: typing.Optional[typing.Text]
    ):
        _on_error = self._xml_and_get_on_error(doc, tag, text, on_next, on_error)

        attributes = {
            'name': self.identifier(),
        }
        if self.__credential:
            attributes['cred'] = self.__credential
        if self.__retry_max is not None:
            attributes['retry-max'] = str(self.__retry_max)
        if self.__retry_interval is not None:
            attributes['retry-interval'] = str(self.__retry_interval)
        with tag(self.xml_tag(), **attributes):
            self.__action._xml(doc, tag, text)
            doc.stag('ok', to=on_next)
            doc.stag('error', to=_on_error)

        return doc


class Decision(_AbstractWorkflowEntity):
    """Node specifying a switch/case."""

    def __init__(
            self,
            default,        # type: _AbstractWorkflowEntity
            choices,        # type: typing.Dict[typing.Text, _AbstractWorkflowEntity]
            name=None,      # type: typing.Optional[typing.Text]
            on_error=None,  # type: typing.Optional[_AbstractWorkflowEntity]
    ):
        # type: (...) -> None
        super(Decision, self).__init__(xml_tag='decision', name=name, on_error=on_error)
        assert default, 'A default must be supplied'
        assert choices, 'At least one choice required'
        self.__default = copy.deepcopy(default)
        self.__choices = copy.deepcopy(choices)

    def _xml(
            self,
            doc,        # type: yattag.doc.Doc
            tag,        # type: yattag.doc.Doc.tag
            text,       # type: yattag.doc.Doc.Text
            on_next,    # type: typing.Text
            on_error    # type: typing.Optional[typing.Text]
    ):
        _on_error = self._xml_and_get_on_error(doc, tag, text, on_next, on_error)

        # Write switch/case
        with tag(self.xml_tag(), name=self.identifier()):
            with tag('switch'):
                for case, dest in self.__choices.items():
                    with tag('case', to=dest.identifier()):
                        doc.text(case)
                doc.stag('default', to=self.__default.identifier())

        # Write contained actions
        self.__default._xml(doc, tag, text, on_next, _on_error)
        for choice in self.__choices.values():
            choice._xml(doc, tag, text, on_next, _on_error)

        return doc

    def __iter__(self):
        # type: () -> typing.Generator[_AbstractWorkflowEntity, None, None]
        for action in self.__choices.values():
            yield action
        yield self.__default
        for action in super(Decision, self).__iter__():
            yield action


class Serial(_AbstractWorkflowEntity):
    """Sequence of entities to execute (implemented by chaining entities and 'OK' transitions)"""

    def __init__(self, *entities, **kwargs):
        # type: (*_AbstractWorkflowEntity, **_AbstractWorkflowEntity) -> None
        super(Serial, self).__init__(xml_tag='unknown', on_error=kwargs.get(str('on_error')))
        self.__entities = tuple(copy.deepcopy(entities))  # type: typing.Tuple[_AbstractWorkflowEntity, ...]

    def identifier(self):  # type: () -> typing.Text
        return self.__entities[0].identifier()

    def _xml(
            self,
            doc,        # type: yattag.doc.Doc
            tag,        # type: yattag.doc.Doc.tag
            text,       # type: yattag.doc.Doc.Text
            on_next,    # type: typing.Text
            on_error    # type: typing.Optional[typing.Text]
    ):
        _on_error = self._xml_and_get_on_error(doc, tag, text, on_next, on_error)
        entity_nextidentifier = zip(self.__entities, itertools.chain(
            (a.identifier() for a in self.__entities[1:]),
            (on_next,)
        ))
        for entity, next_identifier in entity_nextidentifier:
            entity._xml(doc, tag, text, on_next=next_identifier, on_error=_on_error)
        return doc

    def __iter__(self):
        # type: () -> typing.Generator[_AbstractWorkflowEntity, None, None]
        for entity in self.__entities:
            yield entity
        for entity in super(Serial, self).__iter__():
            if entity is not self:  # Don't return self because this collection doesn't result in a node
                yield entity


class Parallel(_AbstractWorkflowEntity):
    """Set of entities to execute in parallel (implemented as fork/join tag pair)"""

    def __init__(self, *entities, **kwargs):
        # type: (*_AbstractWorkflowEntity, **typing.Union[_AbstractWorkflowEntity, typing.Text]) -> None
        name = kwargs.get(str('name'))
        name = name if isinstance(name, six.string_types) else None
        on_error = kwargs.get(str('on_error'))
        on_error = on_error if isinstance(on_error, _AbstractWorkflowEntity) else None
        super(Parallel, self).__init__(
            xml_tag='fork',
            name=name,
            on_error=on_error
        )
        assert entities, 'At least 1 entity required'
        self.__entities = frozenset(copy.deepcopy(entities))  # type: typing.FrozenSet[_AbstractWorkflowEntity]

    def _xml(
            self,
            doc,        # type: yattag.doc.Doc
            tag,        # type: yattag.doc.Doc.tag
            text,       # type: yattag.doc.Doc.Text
            on_next,    # type: typing.Text
            on_error    # type: typing.Optional[typing.Text]
    ):
        _on_error = self._xml_and_get_on_error(doc, tag, text, on_next, on_error)
        with tag(self.xml_tag(), name=self.identifier()):
            for entity in self.__entities:
                doc.stag('path', start=entity.identifier())
        for entity in self.__entities:
            entity._xml(doc, tag, text, on_next=self.create_identifier('join'), on_error=_on_error)
        doc.stag('join', name=self.create_identifier('join'), to=on_next)
        return doc

    def __iter__(self):
        # type: () -> typing.Generator[_AbstractWorkflowEntity, None, None]
        for entity in self.__entities:
            yield entity
        for entity in super(Parallel, self).__iter__():
            yield entity


class WorkflowApp(XMLSerializable):

    def __init__(
            self,
            name,                         # type: typing.Text
            parameters=None,              # type: typing.Optional[PropertyValuesType]
            configuration=None,           # type: typing.Optional[PropertyValuesType]
            credentials=None,             # type: typing.Optional[typing.Iterable[Credential]]
            job_tracker=None,             # type: typing.Optional[typing.Text]
            name_node=None,               # type: typing.Optional[typing.Text]
            job_xml_files=None,           # type: typing.Optional[JobXmlFilesType]
            entities=None,                # type: typing.Optional[_AbstractWorkflowEntity]
    ):
        # type: (...) -> None
        XMLSerializable.__init__(self, 'workflow-app')

        # XML-document-related values
        self.__name = validate_xml_name(name)
        self.__parameters = Parameters(parameters)
        self.__global_configuration = GlobalConfiguration(
            job_tracker=job_tracker,
            name_node=name_node,
            job_xml_files=job_xml_files,
            configuration=configuration
        )
        self.__credentials = copy.deepcopy(credentials) or []
        self.__entities = copy.deepcopy(entities) or Serial()
        self.__validate()

    def __validate(self):  # type () -> None
        entity_identifiers = []
        credentials_needed = set()

        def _parse_entity(entity):
            # Each entity's identifier should be unique
            entity_identifiers.append(entity.identifier())
            # If the entity refers to a credential by name, it should be defined upon instantiation
            if hasattr(entity, 'credential'):
                credential = entity.credential()
                if credential:
                    credentials_needed.add(credential)

        # Parse entitys for attributes
        if self.__entities:
            for entity in set(self.__entities):
                _parse_entity(entity)

        # Verify that all needed credentials are defined
        credentials_provided = frozenset([cred.name for cred in self.__credentials])
        assert credentials_needed <= credentials_provided, (
            'Missing credentials: %s' % ', '.join(credentials_needed - credentials_provided)
        )

        # Verify that no duplicate identifiers are used
        duplicate_identifiers = tuple(
            item for item, count in collections.Counter(entity_identifiers).items() if count > 1)
        assert not duplicate_identifiers, 'Name(s) reused: %s' % ', '.join(sorted(duplicate_identifiers))

    def _xml(self, doc, tag, text):
        # type: (yattag.doc.Doc, yattag.doc.Doc.tag, yattag.doc.Doc.text) -> yattag.doc.Doc
        with tag(self.xml_tag, name=self.__name, xmlns="uri:oozie:workflow:0.5"):

            # Preamble
            if self.__parameters:
                self.__parameters._xml(doc, tag, text)
            if self.__global_configuration:
                self.__global_configuration._xml(doc, tag, text)
            if self.__credentials:
                with tag('credentials'):
                    for credential in self.__credentials:
                        credential._xml(doc, tag, text)

            # Create a serial collection of workflow entities to hold entities
            doc.stag('start', to=self.__entities.identifier() if self.__entities else 'end')
            self.__entities._xml(doc, tag, text, on_next='end', on_error=None)
            doc.stag('end', name='end')

        return doc
